# Copyright (C) [2015-] [Thomson Reuters LLC]
# Copyright (C) [2015-] [Panos Kittenis]
# Copyright (C) [2014-2015] [Vimeo, LLC]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing InfluxDBLeafNode class"""

from __future__ import absolute_import
from graphite_api.node import LeafNode

class InfluxDBLeafNode(LeafNode):
    """Tell Graphite-Api that our leaf node supports multi-fetch"""
    __fetch_multi__ = 'influxdb'



# Copyright (C) [2015-] [Thomson Reuters LLC]
# Copyright (C) [2015-] [Panos Kittenis]
# Copyright (C) [2014-2015] [Vimeo, LLC]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Graphite template parsing functions per InfluxDB's Graphite service template syntax"""

from __future__ import absolute_import, print_function
import re
import logging
from collections import deque

from .utils import heapsort
from .constants import GRAPHITE_PATH_REGEX_PATTERN
from .classes.tree import NodeTreeIndex

logger = logging.getLogger('influxgraph')


class InvalidTemplateError(Exception):
    """Raised on Graphite template configuration validation errors"""
    pass

def parse_series(series, all_fields, graphite_templates,
                 separator='.'):
    index = NodeTreeIndex()
    for serie in series:
        # If we have metrics with tags in them split them out and
        # pre-generate a correctly ordered split path for that metric
        # to be inserted into index
        if graphite_templates:
            for split_path in _get_series_with_tags(
                    serie, all_fields, graphite_templates,
                    separator=separator):
                index.insert_split_path(split_path)
        # Series with tags and no templates,
        # add only measurement to index
        elif ',' in serie:
            index.insert(serie.split(',')[0])
        # No tags, no template
        else:
            index.insert(serie)
    return index

def _parse_influxdb_graphite_templates(templates, separator='.'):
    # Logic converted to Python from InfluxDB's Golang Graphite template parsing
    # Format is [filter] <template> [tag1=value1,tag2=value2]
    parsed_templates = []
    for pattern in templates:
        template = pattern
        _filter = ""
        parts = template.split()
        if len(parts) < 1:
            continue
        elif len(parts) >= 2:
            if '=' in parts[1]:
                template = parts[0]
            else:
                _filter = parts[0]
                template = parts[1]
        # Parse out the default tags specific to this template
        default_tags = {}
        if '=' in parts[-1]:
            tags = [d.strip() for d in parts[-1].split(',')]
            for tag in tags:
                tag_items = [d.strip() for d in tag.split('=')]
                default_tags[tag_items[0]] = tag_items[1]
        parsed_templates.append((generate_filter_regex(_filter),
                                 _generate_template_tag_index(template),
                                 default_tags, separator))
    for (_, template, _, _) in parsed_templates:
        _template_sanity_check(template)
    return parsed_templates

def _template_sanity_check(template):
    field = ""
    measurement_wildcard, field_wildcard = False, False
    for tag in template.values():
        if tag == 'measurement*':
            measurement_wildcard = True
        if tag == 'field*':
            field_wildcard = True
        if tag == 'field':
            if field:
                raise InvalidTemplateError(
                    "'field' can only be used once in each template - %s",
                    template)
            field = tag
    if measurement_wildcard and field_wildcard:
        raise InvalidTemplateError(
            "Either 'field*' or 'measurement*' can be used in each template, not both - %s",
            template)

def apply_template(metric_path_parts, template, default_tags, separator='.'):
    """Apply template to metric path parts and return measurements, tags and field"""
    measurement = []
    tags = {}
    field = ""
    for i, tag in template.items():
        if i >= len(metric_path_parts):
            continue
        if tag == 'measurement':
            measurement.append(metric_path_parts[i])
        elif tag == 'field':
            field = metric_path_parts[i]
        elif tag == 'field*':
            field = separator.join(metric_path_parts[i:])
            break
        elif tag == 'measurement*':
            measurement.extend(metric_path_parts[i:])
            break
        elif tag != "":
            tags.setdefault(tag, []).append(metric_path_parts[i])
    for tag, values in tags.items():
        tags[tag] = separator.join(values)
    if default_tags:
        tags.update(default_tags)
    return separator.join(measurement), tags, field

def generate_filter_regex(_filter):
    """Generate compiled regex pattern from filter string"""
    if not _filter:
        return
    return re.compile("^%s" % (_filter.replace('.', r'\.').replace('*', '%s+' % (
        GRAPHITE_PATH_REGEX_PATTERN,))))

def _generate_template_tag_index(template):
    _tags = template.split('.')
    tags = {}
    for i, tag in enumerate(_tags):
        if not tag:
            tag = None
        tags[i] = tag
    return tags

def _get_series_with_tags(serie, all_fields, graphite_templates,
                               separator='.'):
    paths = serie.split(',')
    if not graphite_templates:
        return [paths[0:1]]
    series = deque()
    split_path, template = _split_series_with_tags(paths, graphite_templates)
    if not split_path:
        # No template match
        return series
    if 'field' in template.values() or 'field*' in template.values():
        _add_fields_to_paths(
            all_fields[paths[0]], split_path, series, separator=separator)
    else:
        series.append(split_path)
    return series

def _split_series_with_tags(paths, graphite_templates):
    split_path, template = deque(), None
    tags_values = [p.split('=') for p in paths[1:]]
    for (_, template, _, separator) in graphite_templates:
        _make_path_from_template(
            split_path, paths[0], template, tags_values)
        # Split path should be at least as large as number of wanted
        # template tags taking into account measurement and number of fields
        # in template
        field_inds = len([v for v in template.values()
                          if v and 'field' in v])
        if (len(split_path) + field_inds) >= len(
                [k for k, v in template.items() if v]):
            break
        # Reset path if template does not match
        else:
            split_path = []
    path = [p[1] for p in heapsort(split_path)] if split_path \
           else split_path
    return path, template

def _make_path_from_template(split_path, measurement, template, tags_values,
                             separator='.'):
    measurement_found = False
    if not tags_values and separator in measurement and \
       'measurement*' == [t for t in template.values() if t][0]:
        for i, measurement in enumerate(measurement.split(separator)):
            split_path.append((i, measurement))
        return
    for (tag_key, tag_val) in tags_values:
        for i, tmpl_tag_key in template.items():
            if not tmpl_tag_key:
                continue
            if tag_key == tmpl_tag_key:
                split_path.append((i, tag_val))
            elif 'measurement' in tmpl_tag_key and not measurement_found:
                measurement_found = True
                split_path.append((i, measurement))

def _add_fields_to_paths(fields, split_path, series,
                         separator='.'):
    for field_key in fields:
        field_keys = [f for f in field_key.split(separator)
                      if f != 'value']
        series.append(split_path + field_keys)



# Copyright (C) [2015-] [Thomson Reuters LLC]
# Copyright (C) [2015-] [Panos Kittenis]
# Copyright (C) [2014-2015] [Vimeo, LLC]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tree representation of Graphite metrics"""

from __future__ import absolute_import, print_function
import sys
import json
from collections import deque

from graphite_api.utils import is_pattern
from graphite_api.finders import match_entries

def _encode_bytes(_str):
    if not isinstance(b'', str):
        return _str.encode('utf-8')
    return bytes(_str)

def _decode_str(_str):
    try:
        return _str.decode('utf-8')
    except AttributeError:
        pass
    return _str

class Node(object):
    """Node class of a graphite metric"""
    __slots__ = ('children')
    
    def __init__(self):
        self.children = None

    def is_leaf(self):
        """Returns True/False depending on whether self is a LeafNode or not"""
        return self.children is None

    def insert(self, paths):
        """Insert path in this node's children"""
        if len(paths) == 0:
            return
        if self.children is None:
            self.children = ()
        child_name = paths.popleft()
        for (_child_name, node) in self.children:
            # Fast path for end of recursion - avoids extra recursion
            # for empty paths list
            if len(paths) == 0 and child_name == _child_name:
                return
            elif child_name == _child_name:
                return node.insert(paths)
        node = Node()
        self.children += ((child_name, node),)
        return node.insert(paths)

    def to_array(self):
        """Return list of (name, children) items for this node's children"""
        return [(_decode_str(name), node.to_array(),) for (name, node,) in self.children] \
          if self.children is not None else None

    @staticmethod
    def from_array(array):
        """Load given parent node's children from array"""
        metric = Node()
        if array is None:
            return metric
        else:
            metric.children = ()
        for child_name, child_array in array:
            child = Node.from_array(child_array)
            metric.children += ((_encode_bytes(child_name), child),)
        return metric

class NodeTreeIndex(object):
    """Node tree index class with graphite glob searches per sub-part of a
    query
    """
    __slots__ = ('index')

    def __init__(self):
        self.index = Node()

    def insert_series(self, series):
        for serie in series:
            self.insert(serie)

    def insert(self, metric_path):
        """Insert metric path into tree index"""
        paths = deque([_encode_bytes(s) for s in metric_path.split('.')])
        self.index.insert(paths)

    def insert_split_path(self, paths):
        """Insert already split path into tree index"""
        self.index.insert(deque([_encode_bytes(s) for s in  paths]))

    def clear(self):
        """Clear tree index"""
        self.index.children = None

    def query(self, query):
        """Return nodes matching Graphite glob pattern query"""
        nodes = sorted(self.search(self.index, query.split('.'), []))
        return ({'metric': '.'.join(path), 'is_leaf': node.is_leaf()}
                for path, node in nodes)

    def search(self, node, split_query, split_path):
        """Return matching children for each query part in split query starting
        from given node"""
        sub_query = split_query[0]
        keys = [_decode_str(key) for (key, _) in node.children] \
          if node.children is not None else []
        matched_paths = match_entries(keys, sub_query)
        matched_children = (
            (_decode_str(path), _node)
            for (path, _node) in node.children
            if _decode_str(path) in matched_paths) \
            if node.children is not None and is_pattern(sub_query) \
            else [(sub_query, [n for (k, n) in node.children
                               if _decode_str(k) == sub_query][0])] \
                               if node.children is not None \
                               and sub_query in keys else []
        for child_name, child_node in matched_children:
            child_path = split_path[:]
            child_path.append(child_name)
            child_query = split_query[1:]
            if len(child_query) > 0:
                for sub in self.search(child_node, child_query, child_path):
                    yield sub
            else:
                yield (child_path, child_node)

    def to_file(self, file_h):
        """Dump tree contents to file handle"""
        data = bytes(json.dumps(self.to_array()), 'utf-8') \
          if not isinstance(b'', str) else json.dumps(self.to_array())
        file_h.write(data)

    def to_array(self):
        """Return array representation of tree index"""
        return self.index.to_array()

    @staticmethod
    def from_array(model):
        """Load tree index from array"""
        metric_index = NodeTreeIndex()
        metric_index.index = Node.from_array(model)
        return metric_index

    @staticmethod
    def from_file(file_h):
        """Load tree index from file handle"""
        data = file_h.read().decode('utf-8') \
          if not isinstance(b'', str) else file_h.read()
        index = NodeTreeIndex.from_array(json.loads(data))
        return index



# Copyright (C) [2015-] [Thomson Reuters LLC]
# Copyright (C) [2015-] [Panos Kittenis]
# Copyright (C) [2014-2015] [Vimeo, LLC]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""InfluxGraph utility functions"""

from __future__ import absolute_import, print_function
import datetime
import sys
import re
import hashlib
from heapq import heappush, heappop

from .constants import INFLUXDB_AGGREGATIONS


def calculate_interval(start_time, end_time, deltas=None):
    """Calculates wanted data series interval according to start and end times

    Returns interval in seconds
    :param start_time: Start time in seconds from epoch
    :param end_time: End time in seconds from epoch
    :type start_time: int
    :type end_time: int
    :param deltas: Delta configuration to use. Defaults hardcoded if no
    configuration is provided
    :type deltas: dict(max time range of query in seconds: interval to use in seconds)

    :rtype: int - *Interval in seconds*
    """
    time_delta = end_time - start_time
    deltas = deltas if deltas else {
        # 15 min -> 10s
        900: 10,
        # 30 min -> 30s
        1800: 30,
        # # 1 hour -> 1s
        # 3600 : 1,
        # # 1 day -> 30s
        # 86400 : 30,
        # 3 days -> 1min
        259200: 60,
        # 7 days -> 5min
        604800: 300,
        # 14 days -> 10min
        1209600: 600,
        # 28 days -> 15min
        2419200: 900,
        # 2 months -> 30min
        4838400: 1800,
        # 4 months -> 1hour
        9676800: 3600,
        # 12 months -> 3hours
        31536000: 7200,
        # 4 years -> 12hours
        126144000: 43200,
        }
    for delta in sorted(deltas.keys()):
        if time_delta <= delta:
            return deltas[delta]
    # 1 day default, or if time range > max configured (4 years default max)
    return 86400

def get_retention_policy(interval, retention_policies):
    """Get appropriate retention policy for interval provided

    :param interval: Interval of query in seconds
    :type interval: int
    :param retention_policies: Retention policy configuration
    :type retention_policies: dict(max time range of interval in seconds: retention policy name)

    :rtype: ``str`` or ``None``
    """
    if not retention_policies:
        return
    for retention_interval in sorted(retention_policies.keys()):
        if interval <= retention_interval:
            return retention_policies[retention_interval]
    # In the case that desired interval is beyond configured interval range,
    # return policy for max interval
    return retention_policies[max(sorted(retention_policies.keys()))]

class Query(object):
    """Graphite-API compatible query class"""

    def __init__(self, pattern):
        self.pattern = pattern


class NullStatsd(object):
    """Fake StatsClient compatible class to use when statsd is not configured"""

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        pass

    def timer(self, key, val=None):
        return self

    def timing(self, key, val):
        pass

    def start(self):
        pass

    def stop(self):
        pass

def _compile_aggregation_patterns(aggregation_functions):
    """Compile aggregation function patterns to compiled regex objects"""
    if not aggregation_functions:
        return
    compiled_aggregations = {}
    for pattern in aggregation_functions.keys():
        if not aggregation_functions[pattern] in INFLUXDB_AGGREGATIONS:
            sys.stderr.write("Requested aggregation function '%s' is not a valid InfluxDB "
                             "aggregation function - ignoring..\n" % (
                                 aggregation_functions[pattern],))
            continue
        try:
            compiled_aggregations[re.compile(r'%s' % (pattern,))] = aggregation_functions[pattern]
        except re.error:
            sys.stderr.write("Error compiling regex pattern '%s' - ignoring..\n" % (
                pattern,))
    return compiled_aggregations

def get_aggregation_func(path, aggregation_functions):
    """Lookup aggregation function for path, if any.
    Defaults to 'mean'.

    :param path: Path to lookup
    :type path: str
    :param aggregation_functions: Aggregation function configuration
    :type aggregation_functions: dict(<pattern>: <compiled regex>)
    """
    if not aggregation_functions:
        return 'mean'
    for pattern in aggregation_functions:
        if pattern.search(path):
            return aggregation_functions[pattern]
    return 'mean'

def read_influxdb_values(influxdb_data, paths, fields):
    """Return key -> values dict for values from InfluxDB data"""
    _data = {}
    if not isinstance(influxdb_data, list):
        influxdb_data = [influxdb_data]
    for infl_data in influxdb_data:
        for i in range(len(infl_data.keys())):
            for field in fields:
                infl_key = infl_data.keys()[i][0]
                if field == 'value':
                    metric = paths[i]
                else:
                    try:
                        metric = [p for p in paths
                                  if infl_key in p and field in p][0]
                    except IndexError:
                        continue
                    else:
                        # Delete metric from paths so that subsequent searches
                        # for differing tag values do not overwrite its data
                        del paths[paths.index(metric)]
                _data[metric] = [d[field]
                                 for d in infl_data.get_points(infl_key)]
    return _data

def gen_memcache_pattern_key(pattern):
    """Generate memcache key from pattern"""
    return hashlib.md5(pattern.encode('utf8')).hexdigest()

def gen_memcache_key(start_time, end_time, aggregation_func, paths):
    """Generate memcache key to use to cache request data"""
    start_time_dt, end_time_dt = datetime.datetime.fromtimestamp(float(start_time)), \
      datetime.datetime.fromtimestamp(float(end_time))
    td = end_time_dt - start_time_dt
    delta = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
    key_prefix = hashlib.md5("".join(paths).encode('utf8')).hexdigest()
    return "".join([key_prefix, aggregation_func, str(delta)]).encode('utf8')

# Function as per Python official documentation
def heapsort(iterable):
    h = []
    for value in iterable:
        heappush(h, value)
    return [heappop(h) for _ in range(len(h))]



from __future__ import print_function
import platform
import os
import sys
import versioneer
from setuptools import setup, find_packages, Extension, Distribution as _Distribution
from distutils.errors import CCompilerError
from distutils.errors import DistutilsExecError
from distutils.errors import DistutilsPlatformError
from distutils.command.build_ext import build_ext

cpython = platform.python_implementation() == 'CPython'

try:
    from Cython.Build import cythonize
except ImportError:
    USING_CYTHON = False
else:
    USING_CYTHON = True

ext = 'pyx' if USING_CYTHON else 'c'

extensions = [Extension("influxgraph.ext.classes.tree",
                        ["influxgraph/ext/classes/tree.%s" % (ext,)]),
              Extension("influxgraph.ext.templates",
                        ["influxgraph/ext/templates.%s" % (ext,)]),]

if USING_CYTHON:
    extensions = cythonize(extensions)
    
ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)

class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            raise BuildFailed()
        
cmdclass = versioneer.get_cmdclass()
cmdclass.update({'build_ext': ve_build_ext})

class Distribution(_Distribution):

    def has_ext_modules(self):
        # We want to always claim that we have ext_modules. This will be fine
        # if we don't actually have them (such as on PyPy) because nothing
        # will get built, however we don't want to provide an overally broad
        # Wheel package when building a wheel without C support. This will
        # ensure that Wheel knows to treat us as if the build output is
        # platform specific.
        return True

def run_setup(ext_modules):
    setup(
        name='influxgraph',
        version=versioneer.get_version(),
        cmdclass=cmdclass,
        url='https://github.com/pkittenis/influxgraph',
        license='apache2',
        author='PK, re-write of graphite-influxdb by Dieter Plaetinck',
        author_email='22e889d8@opayq.com',
        description=('InfluxDB storage plugin for Graphite-API'),
        long_description=open('README.rst').read(),
        packages=find_packages('.'),
        zip_safe=False,
        include_package_data=True,
        platforms='any',
        classifiers=[
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
            'Topic :: System :: Monitoring',
            ],
        install_requires=['influxdb>=3.0.0', 'graphite-api>=1.1.2', 'python-memcached'],
        extras_require={
            'statsd' : ['statsd'],
            },
        distclass=Distribution,
        **ext_modules
        )

ext_modules = {'ext_modules': []}

if not cpython:
    run_setup(ext_modules)
    print("WARNING: C extensions are not supported on this platform,",
          "Pure Python build succeeded")
elif os.environ.get('DISABLE_INFLUXGRAPH_CEXT'):
    run_setup(ext_modules)
    print("DISABLE_INFLUXGRAPH_CEXT is set - not building C extension",
          "Pure Python build succeeded")
else:
    ext_modules['ext_modules'] = extensions
    try:
        run_setup(ext_modules)
    except BuildFailed as exc:
        print(exc.cause, "WARNING: The C extension could not be compiled,",
              "building without it which will incur a performance penalty.",
              "Reasons for failure may be printed above. Build will now",
              "retry without C extension")
        run_setup({})
        print("WARNING: C extension could not be compiled.",
              "Pure Python build succeeded")
